/*
 *  drivers/mtd/am5120/am5120nand.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 *  Overview:
 *   This is a device driver for the NAND flash device found on the
 *   Admtek 5120 board which utilizes the NAND flash.
 */

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/mtd/mtd.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <asm/io.h>
#include <asm/am5120/adm5120.h>
#include "am5120nand.h"
/*
 * MTD structure for NAND flash devices on 5120 board
 */
static struct mtd_info *nand_mtd = NULL;


/*
 * Module stuff
 */

static int nand_io_base = KSEG1ADDR(SMEM1_BASE);

MODULE_PARM(nand_io_base, "i");

/* 
 *	hardware specific access to control-lines
*/
static void nand_hwcontrol(struct mtd_info *mtd, int cmd){

    switch(cmd){

	case NAND_CTL_SETCLE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_SET_CLE)) =  0x01; 
		break;
	case NAND_CTL_CLRCLE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_CLR_CLE)) = 0x01; 
		break;

	case NAND_CTL_SETALE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_SET_ALE)) =  0x01; 
		break;
	case NAND_CTL_CLRALE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_CLR_ALE)) = 0x01; 
		break;

	case NAND_CTL_SETNCE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_SET_CEn)) = 0x01; 
		break;
	case NAND_CTL_CLRNCE: 
		(*(volatile unsigned char *) (nand_io_base + NAND_CLR_CEn)) = 0x01; 
		break;
    }
}

/*
 * Main initialization routine
 */
int __init nand_init (void)
{
	struct nand_chip *this;

	/* Allocate memory for MTD device structure and private data */
	nand_mtd = kmalloc (sizeof(struct mtd_info) + sizeof (struct nand_chip),
				GFP_KERNEL);
	if (!nand_mtd) {
		printk ("Unable to allocate ADM5120 NAND MTD device structure.\n");
		return -ENOMEM;
	}
	
	/* Get pointer to private data */
	this = (struct nand_chip *) (&nand_mtd[1]);

	/* Initialize structures */
	memset((char *) nand_mtd, 0, sizeof(struct mtd_info));
	memset((char *) this, 0, sizeof(struct nand_chip));

	/* Link the private data with the MTD structure */
	nand_mtd->priv = this;

	/*
	 * Set GPIO Port E control register so that the pins are configured
	 * to be outputs for controlling the NAND flash.
	 */
	(*(volatile unsigned int *) (0xB1000200)) = 0x80;	/* 8bit NAND flash */
	(*(volatile unsigned int *) (0xB2000064)) = 0x100;	/* enable NAND flash */
	(*(volatile unsigned int *) (0xB2000008)) = 0x1;	/* boot done */
	
	/* Set SPn */ 
	(*(volatile unsigned char *) (nand_io_base + NAND_SET_SPn)) =  0x01; 
	/* Clear WPn */
	(*(volatile unsigned char *) (nand_io_base + NAND_CLR_WPn)) =  0x01; 
	
	/* Set address of NAND IO lines */
	this->IO_ADDR_R = nand_io_base;
	this->IO_ADDR_W = nand_io_base;
	/* Set address of hardware control function */
	this->hwcontrol = nand_hwcontrol;
	this->dev_ready = NULL;
	/* 10 us command delay time */
	this->chip_delay = 10;
	this->eccmode = NAND_ECC_SOFT;

	/* Scan to find existence of the device */
	if (nand_scan (nand_mtd, 1)) {
		printk("Nand scan ERROR\n");
		kfree (nand_mtd);
		return -EIO;
	}
	printk("Nand scan okay\n");

	/* Allocate memory for internal data buffer */
	this->data_buf = kmalloc (sizeof(u_char) * (nand_mtd->oobblock + nand_mtd->oobsize), GFP_KERNEL);
	if (!this->data_buf) {
		printk ("Unable to allocate NAND data buffer for ADM5120.\n");
		kfree (nand_mtd);
		return -ENOMEM;
	}

	/* Register the partitions */
	add_mtd_partitions(nand_mtd, partition_info, NUM_PARTITIONS);

	/* Return happy */
	return 0;
}
module_init(nand_init);

/*
 * Clean up routine
 */
#ifdef MODULE
static void __exit nand_cleanup (void)
{
	struct nand_chip *this = (struct nand_chip *) &nand_mtd[1];

	/* Unregister the device */
	del_mtd_device (nand_mtd);

	/* Free internal data buffer */
	kfree (this->data_buf);
	
	/* Free the MTD device structure */
	kfree (nand_mtd);
}
module_exit(nand_cleanup);
#endif

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Jeanson Hung <jeansonh@admtek.com.tw>");
MODULE_DESCRIPTION("Board-specific glue layer for NAND flash on ADM5120 board");
